"""Provide the built-in extensions."""

from __future__ import annotations

from betty.extension.cotton_candy import _CottonCandy
from betty.extension.demo import _Demo
from betty.extension.deriver import _Deriver
from betty.extension.gramps import _Gramps
from betty.extension.http_api_doc import _HttpApiDoc
from betty.extension.maps import _Maps
from betty.extension.nginx import _Nginx
from betty.extension.privatizer import _Privatizer
from betty.extension.trees import _Trees
from betty.extension.wikipedia import _Wikipedia


class CottonCandy(_CottonCandy):
    pass


class Demo(_Demo):
    pass


class Deriver(_Deriver):
    pass


class Gramps(_Gramps):
    pass


class HttpApiDoc(_HttpApiDoc):
    pass


class Maps(_Maps):
    pass


class Nginx(_Nginx):
    pass


class Privatizer(_Privatizer):
    pass


class Trees(_Trees):
    pass


class Wikipedia(_Wikipedia):
    pass
