'use strict'

import './file' // eslint-disable-line no-unused-vars
import './search' // eslint-disable-line no-unused-vars
import './show' // eslint-disable-line no-unused-vars
import mainStyle from './main.scss' // eslint-disable-line no-unused-vars
