{% set place = place | default(entity) %}
{% set embedded = embedded | default(False) %}
<span>
    {%- set names = place.names | select_dateds(date_context | default(none)) | list -%}
    {%- if names | length == 0 -%}
        {%- set names = place.names -%}
    {%- endif -%}
    {%- set name = names | negotiate_localizeds -%}
    {%- if name is none -%}
        {%- set name = names | first -%}
    {%- endif -%}

    {%- if not embedded -%}
        <a href="{{ place | url }}">
    {%- endif -%}
    <span{% if name.locale is not none %} lang="{{ name.locale }}"{% endif %}>{{ name }}</span>
    {%- if not embedded -%}
        </a>
    {%- endif -%}
</span>
