<figure>
    <img src="{{ file | image(500, 500) | static_url }}"{% if file.description %} alt="{{ file.description }}"{% endif %}>
    {% if file.description %}
        <figcaption>{{ file.description }}</figcaption>
    {% endif %}
</figure>
